<?php
define ("LOG_PROFILE", LOG_DEBUG + 1);

class LOGData implements Iterator {
 var $from;
 var $to; 
 var $prio;	// maximal priority

 var $cur;
 
 var $file;	// currently openned log file
 var $min;	// $from in current file
 var $max;	// $to in current file

 var $fname;	// current log file name 
 var $info;	// current data
 
 var $key;
 
 var $filter;

 function __construct($from, $to, $prio = false, $filter = false) {
    if ($to < $from)
	throw new ADEIException(translate("Invalid time slice is specified: '%d'-'%d'", $from,$to));

    $this->from = $from;
    $this->to = $to;
    $this->prio = $prio;
    $this->filter = $filter;
    
    $this->file = false;
 }

 function OpenFile($date, $min = false, $max = false) {
    global $TMP_PATH;

    $this->fname = LOGGER::GetLogFile($date);

    if ($this->file) {
	$this->CloseFile();
    }

    $this->file = @fopen($this->fname, "r");
    if ($this->file) {
	$this->min = $min;
	$this->max = $max;
    }

    return $this->file;
 }
 
 function CloseFile() {
    if ($this->file) {
	fclose($this->file);
	$this->file = false;
    }
 }
 
 function rewind() {
    $this->CloseFile();
 
    $mod = $this->from%86400;
    if ($mod) {
	$this->OpenFile($this->from, $this->from, $this->to);
	$this->cur = $this->from + 86400 - $mod;
    } else {
	$this->cur = $this->from;
    }
    
    $this->key = -1;
    $this->next();
 }

 function current() {
    return $this->info;
 }

 function key() {
    return $this->key;
 }

 function next() {
    while (true) {
	while (!$this->file) {
	    $next = $this->cur + 86400;
	    if ($next > $this->to) {
		$this->OpenFile($this->cur, false, $this->to);
		$this->cur = $next;
		break;
	    } else {
		$this->OpenFile($this->cur);
		$this->cur = $next;
	    }
	}

	if (!$this->file) return;

	$min = $this->min;
	$max = $this->max;

	while (!feof($this->file)) {
	    $pos = ftell($this->file);
	    if ($pos === false) break;

	    $hdrlen = fread($this->file, 5);
	    if ((feof($this->file))||(strlen($hdrlen) != 5)||($hdrlen<=0)) break;


	    $header = fread($this->file, $hdrlen);
	    if ((feof($this->file))||(strlen($header) != $hdrlen)) break;

	    if (sscanf($header, "%s %d %s %s %s %s %s %s %lu %s %s %d %d", $iso_time, $prio, $setup, $server, $src, $target, $session, $request, $latency, $pid, $client, $len, $msglen) != 13) break;

	    if (($this->prio !== false)&&($prio > $this->prio)) {
		if (fseek($this->file, $len, SEEK_CUR) < 0)
		    break;
		else 
		    continue;
	    }

	    $time = new DateTime($iso_time);

	    if (($min)||($max)) {
		$unix_time = $time->format("U");

		if ($min) {
		    if ($unix_time < $min) {
			if (fseek($this->file, $len, SEEK_CUR) < 0)
			    break;
			else 
			    continue;
		    }
		    else $min = false;
		}

		if ($max) {
		    if ($unix_time > $max) {
			break;
		    }
		}
	    }

	    if ($this->filter) {
		$skip = false;
		if (($this->filter['setup'])&&($setup != $this->filter['setup'])) $skip = true;
		if (($this->filter['server'])&&($server != $this->filter['server'])) $skip = true;
		if (($this->filter['session'])&&($session != $this->filter['session'])) $skip = true;
		if (($this->filter['request'])&&($request != $this->filter['request'])) $skip = true;
		if (($this->filter['source'])&&($src != $this->filter['source'])) $skip = true;
		if (($this->filter['target'])&&($target != $this->filter['target'])) $skip = true;
		if (($this->filter['pid'])&&($pid != $this->filter['pid'])) $skip = true;
		if (($this->filter['client'])&&($client != $this->filter['client'])) $skip = true;

		if ($skip) {
		    if (fseek($this->file, $len, SEEK_CUR) < 0)
			break;
		    else 
			continue;
		}
	    }

	    if ($msglen) {
	        $msg = fread($this->file, $msglen);
		if (feof($this->file)) break;
	    } else {
		$msg = "";
	    }

#	    echo $msg . "<br/>\n";

	    if (fseek($this->file, $pos + LOGGER::SIZE_RECORD_LENGTH + $hdrlen + $len, SEEK_SET) < 0) break;

	    $this->key++;	    
	    $this->info = array(
		'time'=> $time,
		'priority'=> $prio,
		'setup' => $setup,
		'server' => $server,
		'source' => $src,
		'target' => $target,
		'session' => $session,
		'request' => $request,
		'latency' => $latency,
		'pid' => $pid,
		'client' => $client,
		'message' => $msg,
		'logfile' => basename($this->fname),
		'filepos' => $pos
	    );

	    return;
	}

	$this->CloseFile();
    }
 }


 function valid() {
    return $this->file?true:false;
 }
}



class LOGGER {
 var $console;
 var $catch;
 var $catch_console;
 
 var $log_output;
 var $request;		/* Unique request id */
 var $ts;		/* Timestamp, when request processing has started */

 var $lock;

 const SIZE_RECORD_LENGTH = 5;
 const TIMEOUT = 2;     /* timeout to write log file lock (seconds) */

 function __construct() {
    global $LOGGER_LOG_REQUESTS;
    global $LOGGER_LOG_OUTPUT;
    global $ADEI_RUNDIR;
    
    $this->console = false;
    $this->catch = false;
    $this->catch_console = false;
    $this->ts = gettimeofday();
    $this->request = uniqid();

    $this->lock = new LOCK("logger");

    if ($LOGGER_LOG_REQUESTS) {
	$this->LogMessage("New request is received", false, NULL, LOG_DEBUG);
        if ($LOGGER_LOG_OUTPUT) {
	    $source = basename($_SERVER['SCRIPT_NAME'], ".php");
	    if ((preg_match("/(services)\\/$source\.php$/", $_SERVER['SCRIPT_NAME'], $m))||(preg_match("/(services)$/", $ADEI_RUNDIR, $m))) {
		if (!preg_match("/(getdata)/", $source)) {
		    $this->CatchOutput();
		    $this->log_output = true;
		} else {
		    $this->log_output = true;
		}
	    } 
	}
    }
 } 

 function __destruct() {
    global $LOGGER_LOG_REQUESTS;
    global $LOGGER_LOG_OUTPUT;

    if ($this->catch) {
	if ($this->log_output) {
	    $out = ob_get_contents();
	    ob_end_flush();
	    
	    $extra = array(
		'result' => bin2hex($out)
	    );
	    $this->LogMessage("Request processing is finished", false, $extra, LOG_DEBUG);
	} else {
	    $this->LogOutput();
	}
    } else if (($LOGGER_LOG_REQUESTS)&&($LOGGER_LOG_OUTPUT)) {
	$this->LogMessage("Request processing is finished", false, NULL, LOG_DEBUG);
    }
 }

 static function GetLogDir() {
    global $TMP_PATH;
    return "$TMP_PATH/log/";
 }
 
 static function GetLogFile($date) {
    return sprintf("%s/%s.log", LOGGER::GetLogDir(), date("Ymd", $date));
 }

 function GetLogs ($from, $to, $prio = false, $filter = false) {
    return new LOGData($from, $to, $prio, $filter);
 }
 
 function ParseLogRecord($f) {
    $hdrlen = @fread($f, 5);
    if ((feof($f))||(strlen($hdrlen) != 5)||($hdrlen<=0)) return false;

    $header = @fread($f, $hdrlen);
    if ((feof($f))||(strlen($header) != $hdrlen)) return false;

    if (sscanf($header, "%s %d %s %s %s %s %s %s %lu %s %s %d %d", $iso_time, $prio, $setup, $server, $src, $target, $session, $request, $latency, $pid, $client, $len, $msglen) != 13) return false;
	    
    $time = new DateTime($iso_time);

    if ($msglen) {
	$msg = @fread($f, $msglen);
	if (feof($f)) return false;
    } else {
	$msg = "";
    }

	// skip space in the beggining and eol in the end?    
    $text_info = @fread($f, $len - $msglen);
    if (!$text_info) return false;
    
    $info = json_decode(urldecode($text_info), true);
#    $info = json_decode(pack('H*', substr($text_info,1, strlen($text_info)-2)), true);
    $info['time'] = $time;
    $info['priority'] = $prio;
    $info['setup'] = $setup;
    $info['server'] = $server;
    $info['source'] = $src;
    $info['target'] = $target;
    $info['session'] = $session;
    $info['request'] = $request;
    $info['latency'] = $latency;
    $info['pid'] = $pid;
    $info['message'] = $msg;
    $info['client'] = $client;
    
    return $info;
 }
 
 function GetLogInfo($file, $pos) {
    $fname = sprintf("%s/%s", $this->GetLogDir(), basename($file));
    $f = @fopen($fname, "r");

    if (!$f) 
	throw new ADEIException(translate("It is not possible to open the specified log file (%s)", basename($file)));

    if (($pos < 0)||(fseek($f, $pos, SEEK_SET) === false)) {
	fclose($f);
	throw new ADEIException(translate("Could not set specified position (%d) within specified log file (%s)", $pos, basename($file)));
    } 

    $rec = $this->ParseLogRecord($f);

    fclose($f);	    
    
    if (!$rec)
	throw new ADEIException(translate("Can't parse log record at specified position (%d) within log file (%s)", $pos, basename($file)));
#    else if (is_string($rec))
#	throw new ADEIException(translate("%s log record at specified position (%d) within log file (%s)", $rec, $pos, basename($file));
    
    return $rec;
    
 }

 function EnableConsoleLog($priority = true) {
    $this->console = $priority;
 }
 
 
 function CatchOutput($console = true) {
    $this->catch = ob_start();
    $this->catch_console = $console;
    $this->log_output = false;
 }
 
 function LogOutput() {
    $out = ob_get_contents();

    if ($this->catch_console) {
	ob_end_flush();
    } else {
	ob_end_clean();
    }

    $console = $this->console;
    $this->console = false;
    
    if ($out) {
	$lines = preg_split("/(\\n|<br\/>)+/", $out, -1, PREG_SPLIT_NO_EMPTY);
	if ($lines) {
	    foreach ($lines as $line) {
		$this->LogMessage($line);
	    }
	}
    }
    
    $this->console = true;
    
    ob_start();
 }
 
 function SkipOutput() {
    if ($this->catch_console) {
	ob_flush();
    } else {
	ob_clean();
    }
 }
 
 static function EncodeObject($obj) {
    return urlencode(serialize($obj));
 }
 
 static function DecodeObject($str) {
    return unserialize(urldecode($str));
 }
 
 function LogMessage($msg, $obj = NULL, $extra = NULL, $priority = LOG_WARNING, $source = NULL) {
    global $TMP_PATH;
    global $ADEI_SETUP;
    global $ADEI_SESSION;
    global $ADEI_RUNDIR;
    global $LOGGER_STORE_OBJECTS;

    $tod = gettimeofday();

    if ($obj instanceof REQUEST) {
	$req = $obj;
	$location = $obj->GetLocationString();
    } else if ($obj instanceof READER) {
	$req = $obj->req;
	$location = $obj->req->GetLocationString();
    } else if ($obj instanceof CACHE) {
	$req = $obj->req;
	$location = $obj->req->GetLocationString();
    } else if ($obj instanceof EXPORT) {
	$req = $obj->req;
	$location = $obj->req->GetLocationString();
    } else if ($obj instanceof DRAW) {
	$req = $obj->req;
	$location = $obj->req->GetLocationString();
    } else if ($obj === NULL) {
	$req = new REQUEST();
	$location = $req->GetLocationString();
    } else {
	$req = new REQUEST();
	$location = false;
    }

    if (($extra)&&(is_array($extra)))
	$info = $extra;
    else
	$info = array();

    $ts = gettimeofday();

    $latency = ($ts['sec'] - $this->ts['sec'])*1000000 + ($ts['usec'] - $this->ts['usec']);

    if ($priority <= LOG_DEBUG) {
	if ($_SERVER['REQUEST_URI']) {
	    $info['request'] = $_SERVER['REQUEST_URI'];
	} 

	if ($_GET) {
	    $info['GET'] = &$_GET;
	}

	if ($_POST) {
	    $info['POST'] = &$_POST;
	} 

	if ($_SERVER) {
	    $info['REQUEST_DETAILS'] = &$_SERVER;
	} 

	if ($HTTP_RAW_POST_DATA) {
	    $info['RAW_DATA'] = &$HTTP_RAW_POST_DATA;
	}
    }

    if (($LOGGER_STORE_OBJECTS)&&($obj)&&(is_object($obj))) {
	try {
		/* DSToDo: somehow handle PDO exception (not-serializable) and serialize
		everything besides it */
		$info['object'] = $this->EncodeObject($obj);
	} catch (Exception $ex) {
	    if (is_object($obj->req)) {
		try {
		    $info['object'] = $this->DecodeObject($obj->req);
		} catch (Exception $ex2) {}
	    }
	}
    }

    $pid = getmypid();

    if ($ADEI_SETUP) {
	$setup = $ADEI_SETUP;
    } else {
	$setup = "-";
    }

    if (($req)&&($req->props)&&($req->props['db_server'])) {
	$server = $req->props['db_server'];
/*    } else if (($_GET)&&($_GET['db_server'])) {
	$server = $_GET['db_server'];*/
    } else {
	$server = "-";

	if (isset($_SERVER['argv'])) {
	    $pos = array_search("-source", $_SERVER['argv']);
	    if (($pos)&&(($pos + 1) < sizeof($_SERVER['argv']))) {
	        $source_param = $_SERVER['argv'][$pos + 1];
		$source_req = new REQUEST($source_param);
		if ($source_req->props['db_server']) $server = $source_req->props['db_server'];
	    }
	}
    }

    if (!$source) {
	$source = basename($_SERVER['SCRIPT_NAME'], ".php");
	if (preg_match("/services\\/(service).php$/", $_SERVER['SCRIPT_NAME'], $m)) {
	    $source = strtoupper($m[1]) . "(" . $_GET["service"] . ")";
	} else if ((preg_match("/(system|service|admin)\\/$source\.php$/", $_SERVER['SCRIPT_NAME'], $m))||(preg_match("/(system|services|admin)$/", $ADEI_RUNDIR, $m))) {
	    $source = strtoupper($m[1]) . "($source)";
	} else {
	    $source = "ADEI($source)";
	}
    }

    if (($_GET)&&($_GET['target'])) {
	$target = $_GET['target'];
    } else {
	$target = "-";
    }

    if ($ADEI_SESSION) {
	$session = $ADEI_SESSION;
    } else {
	$session = "-";
    }

    if ($_SERVER['HTTP_X_FORWARDED_FOR']) {
        $client = $_SERVER['HTTP_X_FORWARDED_FOR'];
    } elseif ($_SERVER['REMOTE_ADDR']) {
	$client = $_SERVER['REMOTE_ADDR'];
    } else {
	$client = "-";
    }

        // "ip, ip" could be returned
    $client = preg_replace("/\s+/", "", $client);

        // check for 2 equal IP case 
    $clients = preg_split("/,/", $client, 2);
    if ((count($clients) == 2)&&($clients[0] === $clients[1])) $client=$clients[0];

    if ($location)
	$msg = sprintf("%s. %s", $location, $msg);
    
    if (($this->console === true)||($priority <= $this->console)) {
	if ($this->catch) {
	    $this->LogOutput();
	    echo "$msg\n";
	    $this->SkipOutput();
	}
    }

    if (!is_dir("$TMP_PATH/log")) {
	if (!@mkdir("$TMP_PATH/log", 0777, true)) return;
	@chmod("$TMP_PATH/log", 0777);
    }
    
    $usec = str_pad($tod['usec'], 6, "0", STR_PAD_LEFT);

    $msg = preg_replace("/\n/", "[[BR]]", $msg);

    $info_str = urlencode(json_encode($info));
#    $info_str = bin2hex(json_encode($info));
    $infolen = strlen($info_str);

    $msglen = strlen($msg);
    $msglen_str = sprintf("%d", $msglen);
    
    $fulllen = $msglen + $infolen + strlen(" \n");

    adeiSetSystemTimezone();
    $header = sprintf("%s %d %s %s %s %s %s %s %lu %s %s %d %d", date("Y-m-d\TH:i:s." . $usec . "0P", $tod['sec']), $priority, $setup, $server, $source, $target, $session, $this->request, $latency, $pid, $client, $fulllen, $msglen);
    adeiRestoreTimezone();
    
    $header_size = strlen($header) + strlen("  ");
    
    $fname = sprintf("$TMP_PATH/log/%s.log", date("Ymd"));
    
    $umask = @umask(0);

        // File position after "a+" may get messed if multiple instances write the log unless we ensure single access already before openning
    if ($this->lock->TimedLock(LOCK::BLOCK|LOCK::EXCLUSIVE, LOGGER::TIMEOUT)) {
        $f = @fopen($fname, "a+");
        if ($f) {
//              No reason to replicate the lock
//    	    if (LOCK::timed_flock($f, LOCK_EX, LOGGER::TIMEOUT)) {
	        fprintf($f, "%" . LOGGER::SIZE_RECORD_LENGTH . "d %s %s %s\n", $header_size, $header, $msg, $info_str);
//	        flock($f, LOCK_UN);
//	    }
	    fclose($f);
        }
        $this->lock->UnLock();
    }    

    @umask($umask);
 }
 
 function LogException(ADEIException $ae, $msg = NULL, $req = NULL, $extra = NULL, $priority = LOG_CRIT, $source = NULL) {
    global $LOGGER_STORE_OBJECTS;
    
    $einfo = $ae->getInfo();
    
    if ($msg) {
	if (strstr($msg, "%s")) $info = sprintf($msg, $einfo);
	else $info = "$msg: $einfo";
    } else {
	$info = $einfo;
    }
    
    if ($LOGGER_STORE_OBJECTS) {
	try {
	    if (($extra)&&(is_array($extra))) {
		$extra['exception'] = $this->EncodeObject($ae);
	    } else {
		$extra = array(
		    'exception' => $this->EncodeObject($ae)
		);
	    }
	} catch (Exception $e) {
	}
    }
    
    $this->LogMessage($info, $req, $extra, $priority, $source);
 }


    // Do not accumulate failed requests, forget them.
 function AnalyzeLogs(INTERVAL $ivl, array $filter = NULL, array $opts = NULL) {
    $aggregate_groups = array('avg_latencies', 'slow_avg_latencies', 'max_latencies', 'total_requests', 'errors', 'slow_requests', 'avg_parallel_requests', 'max_parallel_requests');

    $stat_tmpl = array(
	'requests' => 0,
	'errors' => 0,
	'slow_requests' => 0,
	'started_requests' => 0,
	'avg_latency' => 0,
	'max_latency' => 0,
	'avg_parallel_requests' => 0,
	'max_parallel_requests' => 0
    );

    $slow_tmpl = array(
	'avg_latency' => 0,
	'avg_parallel_requests' => 0
    );

    $logs = $this->GetLogs($ivl->GetWindowStart(), $ivl->GetWindowEnd(), false, $filter?$filter:false);

    $requests = array();


    if (!$opts) $opts = array();

    if (array_key_exists('slow_threshold', $opts)) {
	$slow_threshold = $opts['slow_threshold'];
    } else {
	$slow_threshold = 1E+100;
    }

    if (array_key_exists('stat_window', $opts)) {
	$stat_window = $opts['stat_window'];
    } else {
	$stat_window = PHP_INT_MAX;
    }

    if (array_key_exists('report', $opts)) {
	$report = $opts['report'];
    } else {
	$report = "stat";
    }

    if (preg_match("/stat/", $report)) $report_stat = true;
    else $report_stat = false;
    if (preg_match("/request/", $report)) $report_req = true;
    else $report_req = false;

    if (array_key_exists('aggregates', $opts)) {
	$aggregates = $opts['aggregates'];
    } else {
	$aggregates = array();
    }

    if (array_key_exists('re_filters', $opts)) {
	$re_filters = $opts['re_filters'];
    } else {
	$re_filters = array();
    }

    $req = array();
    $stats = array();
    $slows = array();

    foreach ($aggregate_groups as $ag) {
	eval("\$$ag = array();");
    }

    
    foreach($logs as $log) {
#	echo ($ivl->GetWindowSize() . " -- " . $log['time']->format('Y-m-d H:i:s') . "\n");
//	$info = adeiGetLogInfo($log['logfile'], $log['filepos']);

	if ($filter) {
	    if (($filter["db_server"])&&($log['server'] != $filter["db_server"])) continue;
	}

	$request = $log['request'];

//	print_r($re_filters);
	$filtered = false;
	foreach ($re_filters as $var => $re) {
	    eval("\$val = $var;");
	    if (!preg_match($re, $val)) {
		$filtered = true;
		break;
	    }
	}
	if ($filtered) continue;

	if (strcmp($log['message'], "Request processing is finished") == 0) {
	    if ($report_req) {
		$r = array(
		    'time' => $log['time'],
		    'setup' => $log['setup'],
		    'server' => $log['server'],
		    'service' => $log['source'] . (isset($log['target'])?('/' . $log['target']):''),
		    'latency' => $log['latency'] / 1000000
		);

		if (is_array($req[$request]['extra']))
		    $r = array_merge($r, $req[$request]['extra']);

		array_push($requests, $r);
	    }

	    if (!$report_stat) continue;

	    $to = $log['time']->format('U.u');
	    if ((is_array($req[$request]['extra']))&&(array_key_exists('time',  $req[$request]['extra'])))
		$from = floor($req[$request]['extra']['time']->format('U.u'));
	    else
		$from = floor($to);

	    $from = $from - $from%$stat_window;
	    $to = ceil($to);

	    for ($pos = $from; ($pos < $to)||($from == $pos); $pos += $stat_window) {
		if (!isset($stats[$pos])) {
		    $stats[$pos] = $stat_tmpl;
		    $slows[$pos] = $slow_tmpl;
		}
		$stat = &$stats[$pos];
		$slow = &$slows[$pos];

		foreach ($aggregate_groups as $ag) {
//		    echo $pos . "!\n";
		    eval("if (!isset(\$${ag}[\$pos])) { \$${ag}[\$pos] = array(); }; \$cur_$ag = &\$${ag}[\$pos];");
		}

		$n_req = sizeof($req);
		$latency =  $log['latency'] / 1000000;

		$n = $stat['requests'];
		$stat['requests']++;
		if ($latency > $stat['max_latency']) $stat['max_latency'] = $latency;
		if ($n_req > $stat['max_parallel_requests']) $stat['max_parallel_requests'] = $n_req;
		$stat['avg_latency'] = ($n * $stat['avg_latency'] + $latency) / ($n + 1);
		$stat['avg_parallel_requests'] = ($n * $stat['avg_parallel_requests'] + $n_req) / ($n + 1);

		if ($latency > $slow_threshold) {
		    $n_slow = $stat['slow_requests'];
		    $stat['slow_requests']++;

		    $slow['avg_latency'] = ($n_slow * $slow['avg_latency'] + $latency) / ($n_slow + 1);
		    $slow['avg_parallel_requests'] = ($n_slow * $slow['avg_parallel_requests'] + $n_req) / ($n_slow + 1);
		}
		
		foreach ($aggregates as $akey => $avar) {
		    eval("\$avalue = $avar;");

		    foreach ($aggregate_groups as $ag) {
			eval("if (!isset(\$cur_${ag}[\$avalue])) { \$cur_${ag}[\$avalue] = 0; }");
		    }

		    $n_req = 0;
		    foreach ($req as &$ri) {
			if ($ri['ag_types'][$akey] == $avalue) $n_req++;
		    }

		    $n = $cur_total_requests[$avalue]++;

		    if ($latency > $cur_max_latencies[$avalue]) $cur_max_latencies[$avalue] = $latency;
		    if ($n_req > $cur_max_parallel_requests[$avalue]) $cur_max_parallel_requests[$avalue] = $n_req;

		    $cur_avg_latencies[$avalue] = ($n * $cur_avg_latencies[$avalue] + $latency) / ($n + 1);
		    $cur_avg_parallel_requests[$avalue] = ($n * $cur_avg_parallel_requests[$avalue] + $n_req) / ($n + 1);

		    if ($latency > $slow_threshold) {
			$n_slow = $cur_slow_requests[$avalue]++;
			$cur_slow_avg_latencies[$avalue] = ($n_slow * $cur_slow_avg_latencies[$avalue] + $latency) / ($n_slow + 1);
		    }
		}
	    }

	    unset($req[$request]);
	} else if (strcmp($log['message'], "New request is received") == 0) {
	    $ag_types = array();
	    foreach ($aggregates as $akey => $avar) {
		eval("\$avalue = $avar;");
		$ag_types[$akey] = $avalue;
	    }

	    $req[$request] = array(
		'extra' => array(
		    'time' => $log['time'],
		    'errors' => 0
		),
		'ag_types' => $ag_types
	    );

	    $from = floor($log['time']->format('U.u'));
	    $from = $from - $from%$stat_window;

	    $stats[$from]['started_requests']++;
	} else if (strstr($log['message'], "Control lock is obtained")) {
//	} else if (strstr($log['message'], "Starting control")) {
	    $req[$request]['extra']['control_lock_latency'] = $log['latency'] / 1000000;
	} else if (strstr($log['message'], "ZEUS control request is prepared")) {
	    $req[$request]['extra']['zeus_dbrd_latency'] = $log['latency'] / 1000000;
	} else if (strstr($log['message'], "ZEUS control socket is connected")) {
	    $req[$request]['extra']['zeus_conn_latency'] = $log['latency'] / 1000000;
	} else if (strstr($log['message'], "ZEUS control response is received")) {
	    $req[$request]['extra']['zeus_resp_latency'] = $log['latency'] / 1000000;
	} else if ($log['priority'] < LOG_WARNING) {
	    if (!isset($req[$request]['extra']['errors']))
		$req[$request]['extra']['errors'] = 1;
	    else
		$req[$request]['extra']['errors']++;

	    $from = floor($log['time']->format('U.u'));
	    $from = $from - $from%$stat_window;

	    $stats[$from]['errors']++;
	    foreach ($aggregates as $akey => $avar) {
		eval("\$avalue = $avar;");
		if (!isset($errors[$from]))
		    $errors[$from] = array();
		$errors[$from][$avalue]++;
	    }
	}
    }

    $res = array(
	'requests' => $requests,
	'stat' => $stats,
	'slow' => $slows
    );

    foreach ($aggregate_groups as $ag) {
	eval("\$res['$ag'] = \$$ag;");
    }

    return $res;
 }
}

$adei_logger = new LOGGER();

function adeiGetLogs ($from, $to, $prio = false, $filter = false) {
    global $adei_logger;
    return $adei_logger->GetLogs($from, $to, $prio, $filter);
}

function adeiAnalyzeLogs(INTERVAL $ivl, array $filter = NULL, array $opts = NULL) {
    global $adei_logger;
    return $adei_logger->AnalyzeLogs($ivl, $filter, $opts);
}

function adeiGetLogInfo($file, $pos) {
    global $adei_logger;
    return $adei_logger->GetLogInfo($file, $pos);
}

function adeiGetLogger() {
    global $adei_logger;
    return $adei_logger;
}

function adeiEnableConsoleLog($priority = true) {
    global $adei_logger;
    $adei_logger->EnableConsoleLog($priority);
}

function adeiLogException(ADEIException $ae, $msg = NULL, $req = NULL, $extra = NULL, $priority = LOG_CRIT, $source = NULL) {
    global $adei_logger;
    $adei_logger->LogException($ae, $msg, $req, $extra, $priority, $source);
}

function adeiLogMessage($msg, $req = NULL, $extra = NULL, $priority = LOG_WARNING, $source = NULL) {
    global $adei_logger;
    $adei_logger->LogMessage($msg, $req, $extra, $priority, $source);
}

function adeiLogOutput() {
    global $adei_logger;
    $adei_logger->CatchOutput();
}

?>